from django.db import models
from picklefield import PickledObjectField


class TypeChoices:
    BOOL = 'bool'
    DICT = 'dict'
    FLOAT = 'float'
    INTEGER = 'integer'
    LIST = 'list'
    STRING = 'string'

    ALL_CHOICES = (
        (BOOL, 'Bool'),
        (DICT, 'Dict'),
        (FLOAT, 'Float'),
        (INTEGER, 'Integer'),
        (LIST, 'List'),
        (STRING, 'String'),
    )


class SettySettings(models.Model):
    name = models.CharField(max_length=255, primary_key=True)
    value = PickledObjectField()
    type = models.CharField(max_length=8, choices=TypeChoices.ALL_CHOICES)
    created_time = models.DateTimeField(auto_now_add=True)
    updated_time = models.DateTimeField(auto_now=True)

    @property
    def value_unpacked(self):
        return str(self.value)

    @value_unpacked.setter
    def value_unpacked(self, value):
        self.value = value

    def __str__(self):
        return '{}={}'.format(self.name, self.value_unpacked)

    class Meta:
        verbose_name = 'Setty Settings'
        verbose_name_plural = 'Setty Settings'



from importlib import import_module

from django.conf import settings

from .exceptions import InvalidConfigurationError


def _load_backend_class():
    backend = getattr(settings, 'SETTY_BACKEND', None)
    if backend is None:
        raise InvalidConfigurationError('The SETTY_BACKEND setting needs to be defined.')

    return getattr(import_module('setty.backend'), backend)()


class Settings:
    """
    Wrapper class used for accessing/updating setty settings
    """
    _backend = _load_backend_class()

    def __getattr__(self, key):
        return self._backend.get(key)

    def __setattr__(self, key, value):
        self._backend.set(key, value)

    def __dir__(self):
        return [setting.name for setting in self._backend.get_all()]



import json
from distutils.util import strtobool

from django import forms
from django.conf import settings
from django.contrib import admin
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from .models import TypeChoices, SettySettings

SERIALIZERS = {
    TypeChoices.BOOL: lambda x: strtobool(x),
    TypeChoices.DICT: lambda x: json.loads(x),
    TypeChoices.FLOAT: lambda x: float(x),
    TypeChoices.INTEGER: lambda x: int(x),
    TypeChoices.LIST: lambda x: json.loads(x),
    TypeChoices.STRING: lambda x: str(x),
}


class SettingsForm(forms.ModelForm):
    value_unpacked = forms.CharField(label='Value',
                                     help_text='The value to store. '
                                               'List and Dict data types should be defined as JSON strings.')

    def clean_value_unpacked(self):
        serializer = SERIALIZERS[self.cleaned_data['type']]
        try:
            serialized_value = serializer(self.cleaned_data['value_unpacked'])
        except Exception as e:
            raise ValidationError('An error occurred storing the value: {}'.format(e))

        return serialized_value

    def save(self, commit=True):
        instance = super().save(commit=False)
        instance.value = self.cleaned_data['value_unpacked']
        instance.save()

        # Reset item in cache if changed in the admin
        if settings.SETTY_BACKEND == 'CacheBackend':
            from setty.backend import CacheBackend
            CacheBackend().set_in_cache(instance.name, instance.value)

        return instance

    class Meta:
        model = SettySettings
        fields = ['name', 'type', 'value_unpacked']
        read_only_fields = ['created_time', 'updated_time']

        help_texts = {
            'name': _('The setting name.'),
            'type': _('The data type of the setting being stored.'),
        }


@admin.register(SettySettings)
class SettyAdmin(admin.ModelAdmin):
    form = SettingsForm
    list_display = ['name', 'type', 'value_unpacked', 'created_time', 'updated_time']
    readonly_fields = ['created_time', 'updated_time']



import logging

from django.conf import settings
from django.core.cache import cache

from .models import SettySettings

logger = logging.getLogger(__name__)

class DatabaseBackend:
    """
    The simple DatabaseBackend is backed by the Django model storing these settings
    """
    def get_all(self):
        return SettySettings.objects.all()

    def get(self, name):
        setting = SettySettings.objects.get(name=name)
        return setting.value

    def set(self, name, value):
        SettySettings.objects.filter(name=name).update(value=value)
        return value


class CacheBackend(DatabaseBackend):
    """
    CacheBackend uses the Django cache setup to cache values instead of accessing the
    database on each get call.
    """
    def get(self, name):
        cache_key = self._make_cache_key(name)
        setting_value = cache.get(cache_key, '__expired__')
        if setting_value != '__expired__':
            logger.debug('From Cache:', setting_value)
            return setting_value
        logger.debug('From Database:', setting_value)
        return self._retrieve_and_cache_setting(name)

    def _retrieve_and_cache_setting(self, name):
        value = super().get(name)
        return self.set(name, value)

    def set(self, name, value):
        super().set(name, value)
        self.set_in_cache(name, value)
        return value

    def set_in_cache(self, name, value):
        cache.set(self._make_cache_key(name), value, getattr(settings, 'SETTY_CACHE_TTL', 3600))

    def load_all_settings_into_cache(self):
        for setting in self.get_all():
            self.set_in_cache(setting.name, setting.value)

    @staticmethod
    def _make_cache_key(name):
        return ':'.join([getattr(settings, 'SETTY_CACHE_PREFIX', '_dyn_settings_'), name])



import setty


def setty_settings(request):
    """
    Adds the Setty settings config into theRequestContext.

    Add 'setty.context_processors.setty_settings' to the
    TEMPLATE_CONTEXT_PROCESSORS setting to ensure this can be used.
    """
    return {'setty': setty.config}



from setuptools import setup

setup(
    name='django-setty',

    version='1.0.1',

    author='Michael England',
    author_email='michael.k.england@gmail.com',

    license='Apache License Version 2.0',

    packages=['setty'],
    include_package_data=True,

    url='https://github.com/mikeengland/django-setty',

    description='Django app allowing users to configure settings dynamically in the Admin screen',
    long_description='Django Setty is a Django app which allows users to configure settings dynamically '
                     'in the Django Admin console. A number of data types are supported - dicts, lists, '
                     'integers, floats, strings and booleans. Two backends are bundled - one is a database backend '
                     'which uses the database to retrieve the settings, and the other uses Memcached to cache the '
                     'settings defined in the database.',

    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
    ],
    keywords='django dynamic live settings setty django-setty admin cache',

    install_requires=['Django', 'django-picklefield', 'python-memcached'],
    test_suite='setty.tests',
)
